#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "caml/mlvalues.h"
#include "caml/alloc.h"
#include "caml/memory.h"
#include "caml/misc.h"
#include "caml/custom.h"
#include "caml/fail.h"
#include "caml/unixsupport.h"
#include "caml/callback.h"
#include "caml/bigarray.h"
#include "caml/threads.h"

static unsigned int uint_val(value v);
static value protected_copy_string(const char *s);

/* This file is included into netgss_bindings_stubs.c */

#ifdef __APPLE__
#include <GSS/gssapi.h>
#else
#ifdef NETGSS_VARIANT_INCLUDE_GSS
#include <gss.h>
#else
#include <gssapi.h>
#endif
#endif

/* In the following tag=0 means that the value was allocated by the GSSAPI
   provider, and needs to be deallocated by the special GSSAPI function for
   this purpose. tag=1 means that we allocated the value, and are also
   responsible for freeing it. tag=2 means something special.
*/

static value twrap_gss_buffer_t(long tag, gss_buffer_t buf);
static value twrap_gss_OID(long tag, gss_OID oid);
static value twrap_gss_OID_set(long tag, gss_OID_set set);
static value  wrap_gss_channel_bindings_t(gss_channel_bindings_t cb);
static value  wrap_gss_ctx_id_t(gss_ctx_id_t ctx);
static value  wrap_gss_cred_id_t(gss_cred_id_t cred);
static value  wrap_gss_name_t(gss_name_t name);
static value  wrap_gss_OID(gss_OID oid);
static value  wrap_gss_OID_set(gss_OID_set set);

static gss_buffer_t unwrap_gss_buffer_t(value);
static gss_OID      unwrap_gss_OID(value);
static gss_OID_set  unwrap_gss_OID_set(value);
static gss_ctx_id_t unwrap_gss_ctx_id_t(value);
static gss_name_t   unwrap_gss_name_t(value);
static gss_cred_id_t unwrap_gss_cred_id_t(value);

static long         tag_gss_buffer_t(value);

static void attach_gss_buffer_t(value, value);


typedef OM_uint32 flags;
typedef int status_type_t;


#define raise_null_pointer netgss_null_pointer

static void netgss_null_pointer(void) {
    caml_raise_constant(*caml_named_value
                          ("Netgss_bindings.Null_pointer"));
}


static void netgss_free_buffer_contents(long tag, gss_buffer_t buf) {
    if (buf->value != NULL) {
        if (tag == 0) {
            OM_uint32 major, minor;
            major = gss_release_buffer(&minor, buf);
            if ((major & 0xffff0000) != 0)
                fprintf(stderr, "Netgss: error from gss_release_buffer\n");
        } else {
            if (tag == 1) {
                stat_free(buf->value);
            }
        }
    }
    buf->value = NULL;
    buf->length = 0;
}


static void netgss_free_buffer(long tag, gss_buffer_t buf) {
    netgss_free_buffer_contents(tag, buf);
    /* The descriptor is always allocated by us: */
    stat_free(buf);
}


static gss_buffer_t netgss_alloc_buffer(void) {
    gss_buffer_t buf;
    buf = (gss_buffer_t) stat_alloc(sizeof(gss_buffer_desc));
    buf->value = NULL;
    buf->length = 0;
    return buf;
}


static void init_gss_buffer_t(gss_buffer_t *p) {
    gss_buffer_t buf;
    buf = netgss_alloc_buffer();
    buf->length = 0;
    buf->value = NULL;
    *p = buf;
}


CAMLprim value netgss_release_buffer(value b) {
    gss_buffer_t buf;
    buf = unwrap_gss_buffer_t(b);
    netgss_free_buffer_contents(tag_gss_buffer_t(b), buf);
    return Val_unit;
}


CAMLprim value netgss_buffer_of_string(value s, value pos, value len) {
    gss_buffer_t buf;
    if (Long_val(len) < 0 || Long_val(pos) < 0 ||
        Long_val(pos) > caml_string_length(s) - Long_val(len))
        invalid_argument("buffer_of_string");
    buf = netgss_alloc_buffer();
    buf->length = Long_val(len);
    buf->value = stat_alloc(buf->length);
    memcpy(buf->value, String_val(s) + Long_val(pos), buf->length);
    return twrap_gss_buffer_t(1, buf);
}


CAMLprim value netgss_buffer_of_memory(value m) {
    value r;
    gss_buffer_t buf;
    buf = netgss_alloc_buffer();
    buf->length = caml_ba_byte_size(Caml_ba_array_val(m));
    buf->value = Caml_ba_data_val(m);
    r = twrap_gss_buffer_t(2, buf);
    attach_gss_buffer_t(r, m);
    return r;
}


CAMLprim value netgss_string_of_buffer(value b) {
    gss_buffer_t buf;
    value s;
    buf = unwrap_gss_buffer_t(b);
    if (buf->value == NULL) {
        s = caml_alloc_string(0);
    } else {
        s = caml_alloc_string(buf->length);
        memcpy(String_val(s), buf->value, buf->length);
    }
    return s;
}


CAMLprim value netgss_memory_of_buffer(value b) {
    gss_buffer_t buf;
    value m;
    buf = unwrap_gss_buffer_t(b);
    if (buf->value == NULL) netgss_null_pointer();
    m = caml_ba_alloc_dims(CAML_BA_UINT8 | CAML_BA_C_LAYOUT, 1,
                           buf->value, buf->length);
    return m;
}


static void netgss_free_oid(long tag, gss_OID buf) {
    if (tag == 0 || buf == GSS_C_NO_OID) {
        /* OIDs from the provider are to be considered as read-only */
    } else {
        stat_free(buf->elements);
        stat_free(buf);
    }
}


static gss_OID netgss_alloc_oid(void) {
    return (gss_OID) stat_alloc(sizeof(gss_OID_desc));
}


static gss_OID netgss_copy_oid(gss_OID buf) {
    gss_OID out;
    if (buf == GSS_C_NO_OID) {
        out = GSS_C_NO_OID;
    } else {
        out = netgss_alloc_oid();
        out->length = buf->length;
        out->elements = stat_alloc(buf->length);
        memcpy(out->elements, buf->elements, buf->length);
    }
    return out;
}


CAMLprim value netgss_oid_of_string(value s) {
    gss_OID buf;
    buf = netgss_alloc_oid();
    buf->length = caml_string_length(s);
    buf->elements = stat_alloc(buf->length);
    memcpy(buf->elements, String_val(s), buf->length);
    return twrap_gss_OID(1, buf);
}


CAMLprim value netgss_string_of_oid(value b) {
    gss_OID buf;
    value s;
    buf = unwrap_gss_OID(b);
    if (buf == GSS_C_NO_OID) 
        caml_raise_not_found();
    s = caml_alloc_string(buf->length);
    memcpy(String_val(s), buf->elements, buf->length);
    return s;
}


static void netgss_free_oid_set(long tag, gss_OID_set set) {
    if (tag == 0 || set == GSS_C_NO_OID_SET) {
        OM_uint32 major, minor;
        major = gss_release_oid_set(&minor, &set);
        if ((major & 0xffff0000) != 0)
            fprintf(stderr, "Netgss: error from gss_release_oid_set\n");
    } else {
        size_t k;
        for (k=0; k < set->count; k++) {
            netgss_free_oid(1, set->elements+k);
        }
        stat_free(set->elements);
        stat_free(set);
    }
}


static gss_OID_set netgss_alloc_oid_set(void) {
    return (gss_OID_set) stat_alloc(sizeof(gss_OID_set_desc));
}


CAMLprim value netgss_array_of_oid_set(value varg) {
    CAMLparam1(varg);
    CAMLlocal2(v1, v2);
    gss_OID_set set;
    size_t k, count;
    set = unwrap_gss_OID_set(varg);
    if (set == GSS_C_NO_OID_SET)
        count = 0;
    else
        count = set->count;
    /* no other way than to always copy the members */
    v1 = caml_alloc(count, 0);
    for (k=0; k<count; k++) {
        v2 = twrap_gss_OID(1, netgss_copy_oid(set->elements+k));
        Store_field(v1, k, v2);
    }
    CAMLreturn(v1);
}


CAMLprim value netgss_oid_set_of_array(value varg) {
    gss_OID_set set;
    gss_OID buf;
    gss_OID *p;
    size_t k;
    value v1;
    if (Wosize_val(varg) == 0)
        return twrap_gss_OID_set(1, GSS_C_NO_OID_SET);
    set = netgss_alloc_oid_set();
    set->count = Wosize_val(varg);
    set->elements = stat_alloc(sizeof(gss_OID) * set->count);
    for (k=0; k<set->count; k++) {
        v1 = Field(varg, k);
        buf = unwrap_gss_OID(v1);
        p = &(set->elements) + k;
        *p = netgss_copy_oid(buf);
    }
    return twrap_gss_OID_set(1, set);
}


static void netgss_free_cred_id(long tag, gss_cred_id_t x) {
    OM_uint32 major, minor;
    if (x != GSS_C_NO_CREDENTIAL) {
        major = gss_release_cred(&minor, &x);
        if ((major & 0xffff0000) != 0)
            fprintf(stderr, "Netgss: error from gss_release_cred\n");
    }
}


static void netgss_free_ctx_id(long tag, gss_ctx_id_t x) {
    OM_uint32 major, minor;
    if (x != GSS_C_NO_CONTEXT) {
        major = gss_delete_sec_context(&minor, &x, GSS_C_NO_BUFFER);
        if ((major & 0xffff0000) != 0)
            fprintf(stderr, "Netgss: error from gss_delete_sec_context\n");
    }
}


static void netgss_free_name(long tag, gss_name_t x) {
    OM_uint32 major, minor;
    if (x != GSS_C_NO_NAME) {
        major = gss_release_name(&minor, &x);
        if ((major & 0xffff0000) != 0)
            fprintf(stderr, "Netgss: error from gss_release_name\n");
    }
}


CAMLprim value netgss_no_cb(value dummy) {
    return wrap_gss_channel_bindings_t(GSS_C_NO_CHANNEL_BINDINGS);
}


CAMLprim value netgss_map_cb(value iaddrty, value iaddr, value aaddrty,
                             value aaddr, value data) {
    gss_channel_bindings_t cb;
    size_t iaddr_len, aaddr_len, data_len;
    iaddr_len = caml_string_length(iaddr);
    aaddr_len = caml_string_length(aaddr);
    data_len = caml_string_length(data);
    cb = (gss_channel_bindings_t)
            stat_alloc(sizeof(struct gss_channel_bindings_struct));
    cb->initiator_addrtype = Int_val(iaddrty);
    cb->initiator_address.length = iaddr_len;
    cb->initiator_address.value = stat_alloc(iaddr_len);
    memcpy(cb->initiator_address.value, String_val(iaddr), iaddr_len);
    cb->acceptor_addrtype = Int_val(aaddrty);
    cb->acceptor_address.length = aaddr_len;
    cb->acceptor_address.value = stat_alloc(aaddr_len);
    memcpy(cb->acceptor_address.value, String_val(aaddr), aaddr_len);
    cb->application_data.length = data_len;
    cb->application_data.value = stat_alloc(data_len);
    memcpy(cb->application_data.value, String_val(data), data_len);
    return wrap_gss_channel_bindings_t(cb);
}


CAMLprim value netgss_no_ctx(value dummy) {
    return wrap_gss_ctx_id_t(GSS_C_NO_CONTEXT);
}

CAMLprim value netgss_is_no_ctx(value context) {
    gss_ctx_id_t ctx;
    ctx = unwrap_gss_ctx_id_t(context);
    return Val_bool(ctx == GSS_C_NO_CONTEXT);
}

CAMLprim value netgss_no_cred(value dummy) {
    return wrap_gss_cred_id_t(GSS_C_NO_CREDENTIAL);
}

CAMLprim value netgss_is_no_cred(value cred) {
    return Val_bool(unwrap_gss_cred_id_t(cred) == GSS_C_NO_CREDENTIAL);
}

CAMLprim value netgss_no_name(value dummy) {
    return wrap_gss_name_t(GSS_C_NO_NAME);
}

CAMLprim value netgss_is_no_name(value name) {
    return Val_bool(unwrap_gss_name_t(name) == GSS_C_NO_NAME);
}

CAMLprim value netgss_indefinite(value dummy) {
    return caml_copy_int32(GSS_C_INDEFINITE);
}

CAMLprim value netgss_no_oid(value dummy) {
    return wrap_gss_OID(GSS_C_NO_OID);
}

CAMLprim value netgss_no_oid_set(value dummy) {
    return wrap_gss_OID_set(GSS_C_NO_OID_SET);
}

static void netgss_free_cb(gss_channel_bindings_t x) {
    if (x != NULL) {
        stat_free(x->initiator_address.value);
        stat_free(x->acceptor_address.value);
        stat_free(x->application_data.value);
    }
}
static unsigned int uint_val(value v) {
    if (Int_val(v) < 0) invalid_argument("negative integer");
    return (unsigned int) Int_val(v);
}

static value protected_copy_string(const char *s) {
    if (s==NULL) raise_null_pointer();
    return caml_copy_string(s);
}

/************** gss_buffer_t *************/

struct absstruct_gss_buffer_t { gss_buffer_t value; long tag; long oid; };

#define absstructptr_gss_buffer_t_val(v) ((struct absstruct_gss_buffer_t *) (Data_custom_val(v)))
#define abs_gss_buffer_t_unwrap(v) (absstructptr_gss_buffer_t_val(v)->value)
#define abs_gss_buffer_t_tag(v) (absstructptr_gss_buffer_t_val(v)->tag)
long abs_gss_buffer_t_oid = 0;

static int abs_gss_buffer_t_compare(value v1, value v2) {
  struct absstruct_gss_buffer_t *p1;
  struct absstruct_gss_buffer_t *p2;
  p1 = absstructptr_gss_buffer_t_val(v1);
  p2 = absstructptr_gss_buffer_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_buffer_t_finalize(value v1) {
  struct absstruct_gss_buffer_t *p1;
  p1 = absstructptr_gss_buffer_t_val(v1);
  netgss_free_buffer(p1->tag, p1->value);
}

static struct custom_operations abs_gss_buffer_t_ops = {
  "",
  abs_gss_buffer_t_finalize,
  abs_gss_buffer_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_buffer_t unwrap_gss_buffer_t(value v) {
  gss_buffer_t r;
  r = abs_gss_buffer_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_gss_buffer_t(value v) {
  return abs_gss_buffer_t_tag(Field(v,0));
}

static value twrap_gss_buffer_t(long tag, gss_buffer_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_gss_buffer_t: NULL pointer");
  v = caml_alloc_custom(&abs_gss_buffer_t_ops, sizeof(struct absstruct_gss_buffer_t), 0, 1);
  absstructptr_gss_buffer_t_val(v)->tag = tag;
  absstructptr_gss_buffer_t_val(v)->value = x;
  absstructptr_gss_buffer_t_val(v)->oid = abs_gss_buffer_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_buffer_t(gss_buffer_t x) {
  return twrap_gss_buffer_t(0, x);
}

static void attach_gss_buffer_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_OID *************/

struct absstruct_gss_OID { gss_OID value; long tag; long oid; };

#define absstructptr_gss_OID_val(v) ((struct absstruct_gss_OID *) (Data_custom_val(v)))
#define abs_gss_OID_unwrap(v) (absstructptr_gss_OID_val(v)->value)
#define abs_gss_OID_tag(v) (absstructptr_gss_OID_val(v)->tag)
long abs_gss_OID_oid = 0;

static int abs_gss_OID_compare(value v1, value v2) {
  struct absstruct_gss_OID *p1;
  struct absstruct_gss_OID *p2;
  p1 = absstructptr_gss_OID_val(v1);
  p2 = absstructptr_gss_OID_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_OID_finalize(value v1) {
  struct absstruct_gss_OID *p1;
  p1 = absstructptr_gss_OID_val(v1);
  netgss_free_oid(p1->tag, p1->value);
}

static struct custom_operations abs_gss_OID_ops = {
  "",
  abs_gss_OID_finalize,
  abs_gss_OID_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_OID unwrap_gss_OID(value v) {
  gss_OID r;
  r = abs_gss_OID_unwrap(Field(v,0));
  return r;
}

static long tag_gss_OID(value v) {
  return abs_gss_OID_tag(Field(v,0));
}

static value twrap_gss_OID(long tag, gss_OID x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_OID_ops, sizeof(struct absstruct_gss_OID), 0, 1);
  absstructptr_gss_OID_val(v)->tag = tag;
  absstructptr_gss_OID_val(v)->value = x;
  absstructptr_gss_OID_val(v)->oid = abs_gss_OID_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_OID(gss_OID x) {
  return twrap_gss_OID(0, x);
}

static void attach_gss_OID(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_OID_set *************/

struct absstruct_gss_OID_set { gss_OID_set value; long tag; long oid; };

#define absstructptr_gss_OID_set_val(v) ((struct absstruct_gss_OID_set *) (Data_custom_val(v)))
#define abs_gss_OID_set_unwrap(v) (absstructptr_gss_OID_set_val(v)->value)
#define abs_gss_OID_set_tag(v) (absstructptr_gss_OID_set_val(v)->tag)
long abs_gss_OID_set_oid = 0;

static int abs_gss_OID_set_compare(value v1, value v2) {
  struct absstruct_gss_OID_set *p1;
  struct absstruct_gss_OID_set *p2;
  p1 = absstructptr_gss_OID_set_val(v1);
  p2 = absstructptr_gss_OID_set_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_OID_set_finalize(value v1) {
  struct absstruct_gss_OID_set *p1;
  p1 = absstructptr_gss_OID_set_val(v1);
  netgss_free_oid_set(p1->tag, p1->value);
}

static struct custom_operations abs_gss_OID_set_ops = {
  "",
  abs_gss_OID_set_finalize,
  abs_gss_OID_set_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_OID_set unwrap_gss_OID_set(value v) {
  gss_OID_set r;
  r = abs_gss_OID_set_unwrap(Field(v,0));
  return r;
}

static long tag_gss_OID_set(value v) {
  return abs_gss_OID_set_tag(Field(v,0));
}

static value twrap_gss_OID_set(long tag, gss_OID_set x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_OID_set_ops, sizeof(struct absstruct_gss_OID_set), 0, 1);
  absstructptr_gss_OID_set_val(v)->tag = tag;
  absstructptr_gss_OID_set_val(v)->value = x;
  absstructptr_gss_OID_set_val(v)->oid = abs_gss_OID_set_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_OID_set(gss_OID_set x) {
  return twrap_gss_OID_set(0, x);
}

static void attach_gss_OID_set(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_cred_id_t *************/

struct absstruct_gss_cred_id_t { gss_cred_id_t value; long tag; long oid; };

#define absstructptr_gss_cred_id_t_val(v) ((struct absstruct_gss_cred_id_t *) (Data_custom_val(v)))
#define abs_gss_cred_id_t_unwrap(v) (absstructptr_gss_cred_id_t_val(v)->value)
#define abs_gss_cred_id_t_tag(v) (absstructptr_gss_cred_id_t_val(v)->tag)
long abs_gss_cred_id_t_oid = 0;

static int abs_gss_cred_id_t_compare(value v1, value v2) {
  struct absstruct_gss_cred_id_t *p1;
  struct absstruct_gss_cred_id_t *p2;
  p1 = absstructptr_gss_cred_id_t_val(v1);
  p2 = absstructptr_gss_cred_id_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_cred_id_t_finalize(value v1) {
  struct absstruct_gss_cred_id_t *p1;
  p1 = absstructptr_gss_cred_id_t_val(v1);
  netgss_free_cred_id(p1->tag, p1->value);
}

static struct custom_operations abs_gss_cred_id_t_ops = {
  "",
  abs_gss_cred_id_t_finalize,
  abs_gss_cred_id_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_cred_id_t unwrap_gss_cred_id_t(value v) {
  gss_cred_id_t r;
  r = abs_gss_cred_id_t_unwrap(Field(v,0));
  return r;
}

static long tag_gss_cred_id_t(value v) {
  return abs_gss_cred_id_t_tag(Field(v,0));
}

static value twrap_gss_cred_id_t(long tag, gss_cred_id_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_cred_id_t_ops, sizeof(struct absstruct_gss_cred_id_t), 0, 1);
  absstructptr_gss_cred_id_t_val(v)->tag = tag;
  absstructptr_gss_cred_id_t_val(v)->value = x;
  absstructptr_gss_cred_id_t_val(v)->oid = abs_gss_cred_id_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_cred_id_t(gss_cred_id_t x) {
  return twrap_gss_cred_id_t(0, x);
}

static void attach_gss_cred_id_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_ctx_id_t *************/

struct absstruct_gss_ctx_id_t { gss_ctx_id_t value; long tag; long oid; };

#define absstructptr_gss_ctx_id_t_val(v) ((struct absstruct_gss_ctx_id_t *) (Data_custom_val(v)))
#define abs_gss_ctx_id_t_unwrap(v) (absstructptr_gss_ctx_id_t_val(v)->value)
#define abs_gss_ctx_id_t_tag(v) (absstructptr_gss_ctx_id_t_val(v)->tag)
long abs_gss_ctx_id_t_oid = 0;

static int abs_gss_ctx_id_t_compare(value v1, value v2) {
  struct absstruct_gss_ctx_id_t *p1;
  struct absstruct_gss_ctx_id_t *p2;
  p1 = absstructptr_gss_ctx_id_t_val(v1);
  p2 = absstructptr_gss_ctx_id_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_ctx_id_t_finalize(value v1) {
  struct absstruct_gss_ctx_id_t *p1;
  p1 = absstructptr_gss_ctx_id_t_val(v1);
  netgss_free_ctx_id(p1->tag, p1->value);
}

static struct custom_operations abs_gss_ctx_id_t_ops = {
  "",
  abs_gss_ctx_id_t_finalize,
  abs_gss_ctx_id_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_ctx_id_t unwrap_gss_ctx_id_t(value v) {
  gss_ctx_id_t r;
  r = abs_gss_ctx_id_t_unwrap(Field(v,0));
  return r;
}

static long tag_gss_ctx_id_t(value v) {
  return abs_gss_ctx_id_t_tag(Field(v,0));
}

static value twrap_gss_ctx_id_t(long tag, gss_ctx_id_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_ctx_id_t_ops, sizeof(struct absstruct_gss_ctx_id_t), 0, 1);
  absstructptr_gss_ctx_id_t_val(v)->tag = tag;
  absstructptr_gss_ctx_id_t_val(v)->value = x;
  absstructptr_gss_ctx_id_t_val(v)->oid = abs_gss_ctx_id_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_ctx_id_t(gss_ctx_id_t x) {
  return twrap_gss_ctx_id_t(0, x);
}

static void set_gss_ctx_id_t(value v, gss_ctx_id_t x) {
  absstructptr_gss_ctx_id_t_val(v)->value = x;
}

static void attach_gss_ctx_id_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_name_t *************/

struct absstruct_gss_name_t { gss_name_t value; long tag; long oid; };

#define absstructptr_gss_name_t_val(v) ((struct absstruct_gss_name_t *) (Data_custom_val(v)))
#define abs_gss_name_t_unwrap(v) (absstructptr_gss_name_t_val(v)->value)
#define abs_gss_name_t_tag(v) (absstructptr_gss_name_t_val(v)->tag)
long abs_gss_name_t_oid = 0;

static int abs_gss_name_t_compare(value v1, value v2) {
  struct absstruct_gss_name_t *p1;
  struct absstruct_gss_name_t *p2;
  p1 = absstructptr_gss_name_t_val(v1);
  p2 = absstructptr_gss_name_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_name_t_finalize(value v1) {
  struct absstruct_gss_name_t *p1;
  p1 = absstructptr_gss_name_t_val(v1);
  netgss_free_name(p1->tag, p1->value);
}

static struct custom_operations abs_gss_name_t_ops = {
  "",
  abs_gss_name_t_finalize,
  abs_gss_name_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_name_t unwrap_gss_name_t(value v) {
  gss_name_t r;
  r = abs_gss_name_t_unwrap(Field(v,0));
  return r;
}

static long tag_gss_name_t(value v) {
  return abs_gss_name_t_tag(Field(v,0));
}

static value twrap_gss_name_t(long tag, gss_name_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_name_t_ops, sizeof(struct absstruct_gss_name_t), 0, 1);
  absstructptr_gss_name_t_val(v)->tag = tag;
  absstructptr_gss_name_t_val(v)->value = x;
  absstructptr_gss_name_t_val(v)->oid = abs_gss_name_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_name_t(gss_name_t x) {
  return twrap_gss_name_t(0, x);
}

static void attach_gss_name_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_channel_bindings_t *************/

struct absstruct_gss_channel_bindings_t { gss_channel_bindings_t value; long tag; long oid; };

#define absstructptr_gss_channel_bindings_t_val(v) ((struct absstruct_gss_channel_bindings_t *) (Data_custom_val(v)))
#define abs_gss_channel_bindings_t_unwrap(v) (absstructptr_gss_channel_bindings_t_val(v)->value)
#define abs_gss_channel_bindings_t_tag(v) (absstructptr_gss_channel_bindings_t_val(v)->tag)
long abs_gss_channel_bindings_t_oid = 0;

static int abs_gss_channel_bindings_t_compare(value v1, value v2) {
  struct absstruct_gss_channel_bindings_t *p1;
  struct absstruct_gss_channel_bindings_t *p2;
  p1 = absstructptr_gss_channel_bindings_t_val(v1);
  p2 = absstructptr_gss_channel_bindings_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_gss_channel_bindings_t_finalize(value v1) {
  struct absstruct_gss_channel_bindings_t *p1;
  p1 = absstructptr_gss_channel_bindings_t_val(v1);
  netgss_free_cb(p1->value);
}

static struct custom_operations abs_gss_channel_bindings_t_ops = {
  "",
  abs_gss_channel_bindings_t_finalize,
  abs_gss_channel_bindings_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static gss_channel_bindings_t unwrap_gss_channel_bindings_t(value v) {
  gss_channel_bindings_t r;
  r = abs_gss_channel_bindings_t_unwrap(Field(v,0));
  return r;
}

static long tag_gss_channel_bindings_t(value v) {
  return abs_gss_channel_bindings_t_tag(Field(v,0));
}

static value twrap_gss_channel_bindings_t(long tag, gss_channel_bindings_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  v = caml_alloc_custom(&abs_gss_channel_bindings_t_ops, sizeof(struct absstruct_gss_channel_bindings_t), 0, 1);
  absstructptr_gss_channel_bindings_t_val(v)->tag = tag;
  absstructptr_gss_channel_bindings_t_val(v)->value = x;
  absstructptr_gss_channel_bindings_t_val(v)->oid = abs_gss_channel_bindings_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_gss_channel_bindings_t(gss_channel_bindings_t x) {
  return twrap_gss_channel_bindings_t(0, x);
}

static void attach_gss_channel_bindings_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** gss_cred_usage_t *************/

static value wrap_gss_cred_usage_t(gss_cred_usage_t x) {
  switch (x) {
    case GSS_C_BOTH: return Val_long(737457313);
    case GSS_C_INITIATE: return Val_long(960571001);
    case GSS_C_ACCEPT: return Val_long(1067124520);
    default: break;
  };
  failwith("wrap_gss_cred_usage_t: unexpected value");
}

static gss_cred_usage_t unwrap_gss_cred_usage_t(value v) {
  switch (Long_val(v)) {
    case 737457313: return GSS_C_BOTH;
    case 960571001: return GSS_C_INITIATE;
    case 1067124520: return GSS_C_ACCEPT;
    default: invalid_argument("unwrap_gss_cred_usage_t");
  };
  failwith("unwrap_gss_cred_usage_t: unexpected value");
}

/************** flags *************/

static value wrap_flags(flags x) {
  CAMLparam0();
  CAMLlocal2(v,u);
  v = Val_long(0);
  if (x & GSS_C_DELEG_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(-69757538);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_MUTUAL_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(-983767337);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_REPLAY_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(690326244);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_SEQUENCE_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(213388682);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_CONF_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(528282375);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_INTEG_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(-245616902);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_ANON_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(483920031);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_PROT_READY_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(947138816);
    Field(u, 1) = v;
    v = u;
  };
  if (x & GSS_C_TRANS_FLAG) {
    u = caml_alloc(2,0);
    Field(u, 0) = Val_long(69206659);
    Field(u, 1) = v;
    v = u;
  };
  CAMLreturn(v);
}

static flags unwrap_flags(value v) {
  flags x = 0;
  while (Is_block(v)) {
    switch (Long_val(Field(v,0))) {
      case -69757538: x |= GSS_C_DELEG_FLAG; break;
      case -983767337: x |= GSS_C_MUTUAL_FLAG; break;
      case 690326244: x |= GSS_C_REPLAY_FLAG; break;
      case 213388682: x |= GSS_C_SEQUENCE_FLAG; break;
      case 528282375: x |= GSS_C_CONF_FLAG; break;
      case -245616902: x |= GSS_C_INTEG_FLAG; break;
      case 483920031: x |= GSS_C_ANON_FLAG; break;
      case 947138816: x |= GSS_C_PROT_READY_FLAG; break;
      case 69206659: x |= GSS_C_TRANS_FLAG; break;
    };
    v = Field(v,1);
  };
  return x;
}

/************** gss_qop_t *************/

static value wrap_gss_qop_t(gss_qop_t x) {
  switch (x) {
    case GSS_C_QOP_DEFAULT: return Val_long(-977007628);
    default: break;
  };
  failwith("wrap_gss_qop_t: unexpected value");
}

static gss_qop_t unwrap_gss_qop_t(value v) {
  switch (Long_val(v)) {
    case -977007628: return GSS_C_QOP_DEFAULT;
    default: invalid_argument("unwrap_gss_qop_t");
  };
  failwith("unwrap_gss_qop_t: unexpected value");
}

/************** status_type_t *************/

static value wrap_status_type_t(status_type_t x) {
  switch (x) {
    case GSS_C_GSS_CODE: return Val_long(-83735803);
    case GSS_C_MECH_CODE: return Val_long(574872527);
    default: break;
  };
  failwith("wrap_status_type_t: unexpected value");
}

static status_type_t unwrap_status_type_t(value v) {
  switch (Long_val(v)) {
    case -83735803: return GSS_C_GSS_CODE;
    case 574872527: return GSS_C_MECH_CODE;
    default: invalid_argument("unwrap_status_type_t");
  };
  failwith("unwrap_status_type_t: unexpected value");
}

value net_gss_accept_sec_context(value context,value acceptor_cred,value input_token,value chan_bindings) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_cred_id_t acceptor_cred__c;
  gss_buffer_t input_token__c;
  gss_channel_bindings_t chan_bindings__c;
  gss_name_t src_name__c;
  gss_OID mech_type__c;
  gss_buffer_t output_token__c;
  flags ret_flags__c;
  OM_uint32 time_rec__c;
  gss_cred_id_t delegated_cred__c;
  OM_uint32 result__c;
  CAMLparam4(context,acceptor_cred,input_token,chan_bindings);
  CAMLlocal5(minor_status,src_name,mech_type,output_token,ret_flags);
  CAMLlocal4(time_rec,delegated_cred,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  acceptor_cred__c = unwrap_gss_cred_id_t(acceptor_cred);
  input_token__c = unwrap_gss_buffer_t(input_token);
  chan_bindings__c = unwrap_gss_channel_bindings_t(chan_bindings);
  init_gss_buffer_t(&output_token__c);
  caml_enter_blocking_section();
  result__c = gss_accept_sec_context(&minor_status__c,&context__c,acceptor_cred__c,input_token__c,chan_bindings__c,&src_name__c,&mech_type__c,output_token__c,&ret_flags__c,&time_rec__c,&delegated_cred__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  context = wrap_gss_ctx_id_t(context__c);
  src_name = wrap_gss_name_t(src_name__c);
  mech_type = wrap_gss_OID(mech_type__c);
  output_token = wrap_gss_buffer_t(output_token__c);
  ret_flags = wrap_flags(ret_flags__c);
  time_rec = caml_copy_int32(time_rec__c);
  delegated_cred = wrap_gss_cred_id_t(delegated_cred__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(9,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = context;
  Field(local_0, 3) = src_name;
  Field(local_0, 4) = mech_type;
  Field(local_0, 5) = output_token;
  Field(local_0, 6) = ret_flags;
  Field(local_0, 7) = time_rec;
  Field(local_0, 8) = delegated_cred;
  CAMLreturn(local_0);
}

value net_gss_acquire_cred(value desired_name,value time_req,value desired_mechs,value cred_usage) {
  OM_uint32 minor_status__c;
  gss_name_t desired_name__c;
  OM_uint32 time_req__c;
  gss_OID_set desired_mechs__c;
  gss_cred_usage_t cred_usage__c;
  gss_cred_id_t cred__c;
  gss_OID_set actual_mechs__c;
  OM_uint32 time_rec__c;
  OM_uint32 result__c;
  CAMLparam4(desired_name,time_req,desired_mechs,cred_usage);
  CAMLlocal5(minor_status,cred,actual_mechs,time_rec,result);
  CAMLlocal1(local_0);
  desired_name__c = unwrap_gss_name_t(desired_name);
  time_req__c = Int32_val(time_req);
  desired_mechs__c = unwrap_gss_OID_set(desired_mechs);
  cred_usage__c = unwrap_gss_cred_usage_t(cred_usage);
  caml_enter_blocking_section();
  result__c = gss_acquire_cred(&minor_status__c,desired_name__c,time_req__c,desired_mechs__c,cred_usage__c,&cred__c,&actual_mechs__c,&time_rec__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  cred = wrap_gss_cred_id_t(cred__c);
  actual_mechs = wrap_gss_OID_set(actual_mechs__c);
  time_rec = caml_copy_int32(time_rec__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(5,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = cred;
  Field(local_0, 3) = actual_mechs;
  Field(local_0, 4) = time_rec;
  CAMLreturn(local_0);
}

value net_gss_init_sec_context(value initiator_cred,value context,value target_name,value mech_type,value req_flags,value time_req,value chan_bindings,value input_token) {
  OM_uint32 minor_status__c;
  gss_cred_id_t initiator_cred__c;
  gss_ctx_id_t context__c;
  gss_name_t target_name__c;
  gss_OID mech_type__c;
  flags req_flags__c;
  OM_uint32 time_req__c;
  gss_channel_bindings_t chan_bindings__c;
  gss_buffer_t input_token__c;
  gss_OID actual_mech_type__c;
  gss_buffer_t output_token__c;
  flags ret_flags__c;
  OM_uint32 time_rec__c;
  OM_uint32 result__c;
  CAMLparam5(initiator_cred,context,target_name,mech_type,req_flags);
  CAMLxparam3(time_req,chan_bindings,input_token);
  CAMLlocal5(minor_status,actual_mech_type,output_token,ret_flags,time_rec);
  CAMLlocal2(result,local_0);
  initiator_cred__c = unwrap_gss_cred_id_t(initiator_cred);
  context__c = unwrap_gss_ctx_id_t(context);
  target_name__c = unwrap_gss_name_t(target_name);
  mech_type__c = unwrap_gss_OID(mech_type);
  req_flags__c = unwrap_flags(req_flags);
  time_req__c = Int32_val(time_req);
  chan_bindings__c = unwrap_gss_channel_bindings_t(chan_bindings);
  input_token__c = unwrap_gss_buffer_t(input_token);
  init_gss_buffer_t(&output_token__c);
  caml_enter_blocking_section();
  result__c = gss_init_sec_context(&minor_status__c,initiator_cred__c,&context__c,target_name__c,mech_type__c,req_flags__c,time_req__c,chan_bindings__c,input_token__c,&actual_mech_type__c,output_token__c,&ret_flags__c,&time_rec__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  context = wrap_gss_ctx_id_t(context__c);
  actual_mech_type = wrap_gss_OID(actual_mech_type__c);
  output_token = wrap_gss_buffer_t(output_token__c);
  ret_flags = wrap_flags(ret_flags__c);
  time_rec = caml_copy_int32(time_rec__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(7,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = context;
  Field(local_0, 3) = actual_mech_type;
  Field(local_0, 4) = output_token;
  Field(local_0, 5) = ret_flags;
  Field(local_0, 6) = time_rec;
  CAMLreturn(local_0);
}

value net_gss_init_sec_context__byte(value * argv, int argn) {
  return net_gss_init_sec_context(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5],argv[6],argv[7]);
}

value net_gss_process_context_token(value context,value token) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_buffer_t token__c;
  OM_uint32 result__c;
  CAMLparam2(context,token);
  CAMLlocal3(minor_status,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  token__c = unwrap_gss_buffer_t(token);
  caml_enter_blocking_section();
  result__c = gss_process_context_token(&minor_status__c,context__c,token__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(2,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  CAMLreturn(local_0);
}

value net_gss_context_time(value context) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  OM_uint32 time_rec__c;
  OM_uint32 result__c;
  CAMLparam1(context);
  CAMLlocal4(minor_status,time_rec,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  result__c = gss_context_time(&minor_status__c,context__c,&time_rec__c);
  minor_status = caml_copy_int32(minor_status__c);
  time_rec = caml_copy_int32(time_rec__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = time_rec;
  CAMLreturn(local_0);
}

value net_gss_get_mic(value context,value qop_req,value message) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  OM_uint32 qop_req__c;
  gss_buffer_t message__c;
  gss_buffer_t msg_token__c;
  OM_uint32 result__c;
  CAMLparam3(context,qop_req,message);
  CAMLlocal4(minor_status,msg_token,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  qop_req__c = Int32_val(qop_req);
  message__c = unwrap_gss_buffer_t(message);
  init_gss_buffer_t(&msg_token__c);
  caml_enter_blocking_section();
  result__c = gss_get_mic(&minor_status__c,context__c,qop_req__c,message__c,msg_token__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  msg_token = wrap_gss_buffer_t(msg_token__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = msg_token;
  CAMLreturn(local_0);
}

value net_gss_verify_mic(value context,value message,value token) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_buffer_t message__c;
  gss_buffer_t token__c;
  OM_uint32 qop_state__c;
  OM_uint32 result__c;
  CAMLparam3(context,message,token);
  CAMLlocal4(minor_status,qop_state,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  message__c = unwrap_gss_buffer_t(message);
  token__c = unwrap_gss_buffer_t(token);
  caml_enter_blocking_section();
  result__c = gss_verify_mic(&minor_status__c,context__c,message__c,token__c,&qop_state__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  qop_state = caml_copy_int32(qop_state__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = qop_state;
  CAMLreturn(local_0);
}

value net_gss_wrap(value context,value conf_req,value qop_req,value input_message) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  int conf_req__c;
  OM_uint32 qop_req__c;
  gss_buffer_t input_message__c;
  int conf_state__c;
  gss_buffer_t output_message__c;
  OM_uint32 result__c;
  CAMLparam4(context,conf_req,qop_req,input_message);
  CAMLlocal5(minor_status,conf_state,output_message,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  conf_req__c = Bool_val(conf_req);
  qop_req__c = Int32_val(qop_req);
  input_message__c = unwrap_gss_buffer_t(input_message);
  init_gss_buffer_t(&output_message__c);
  caml_enter_blocking_section();
  result__c = gss_wrap(&minor_status__c,context__c,conf_req__c,qop_req__c,input_message__c,&conf_state__c,output_message__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  conf_state = Val_bool(conf_state__c);
  output_message = wrap_gss_buffer_t(output_message__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(4,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = conf_state;
  Field(local_0, 3) = output_message;
  CAMLreturn(local_0);
}

value net_gss_unwrap(value context,value input_message) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_buffer_t input_message__c;
  gss_buffer_t output_message__c;
  int conf_state__c;
  OM_uint32 qop_state__c;
  OM_uint32 result__c;
  CAMLparam2(context,input_message);
  CAMLlocal5(minor_status,output_message,conf_state,qop_state,result);
  CAMLlocal1(local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  input_message__c = unwrap_gss_buffer_t(input_message);
  init_gss_buffer_t(&output_message__c);
  caml_enter_blocking_section();
  result__c = gss_unwrap(&minor_status__c,context__c,input_message__c,output_message__c,&conf_state__c,&qop_state__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  output_message = wrap_gss_buffer_t(output_message__c);
  conf_state = Val_bool(conf_state__c);
  qop_state = caml_copy_int32(qop_state__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(5,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = output_message;
  Field(local_0, 3) = conf_state;
  Field(local_0, 4) = qop_state;
  CAMLreturn(local_0);
}

value net_gss_display_status(value status_value,value status_type,value mech_type,value message_context) {
  OM_uint32 minor_status__c;
  OM_uint32 status_value__c;
  status_type_t status_type__c;
  gss_OID mech_type__c;
  OM_uint32 message_context__c;
  gss_buffer_t status_string__c;
  OM_uint32 result__c;
  CAMLparam4(status_value,status_type,mech_type,message_context);
  CAMLlocal4(minor_status,status_string,result,local_0);
  status_value__c = Int32_val(status_value);
  status_type__c = unwrap_status_type_t(status_type);
  mech_type__c = unwrap_gss_OID(mech_type);
  message_context__c = Int32_val(message_context);
  init_gss_buffer_t(&status_string__c);
  result__c = gss_display_status(&minor_status__c,status_value__c,status_type__c,mech_type__c,&message_context__c,status_string__c);
  minor_status = caml_copy_int32(minor_status__c);
  message_context = caml_copy_int32(message_context__c);
  status_string = wrap_gss_buffer_t(status_string__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(4,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = message_context;
  Field(local_0, 3) = status_string;
  CAMLreturn(local_0);
}

value net_gss_indicate_mechs(value dummy) {
  OM_uint32 minor_status__c;
  gss_OID_set mech_set__c;
  OM_uint32 result__c;
  CAMLparam1(dummy);
  CAMLlocal4(minor_status,mech_set,result,local_0);
  result__c = gss_indicate_mechs(&minor_status__c,&mech_set__c);
  minor_status = caml_copy_int32(minor_status__c);
  mech_set = wrap_gss_OID_set(mech_set__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = mech_set;
  CAMLreturn(local_0);
}

value net_gss_compare_name(value name1,value name2) {
  OM_uint32 minor_status__c;
  gss_name_t name1__c;
  gss_name_t name2__c;
  int name_equal__c;
  OM_uint32 result__c;
  CAMLparam2(name1,name2);
  CAMLlocal4(minor_status,name_equal,result,local_0);
  name1__c = unwrap_gss_name_t(name1);
  name2__c = unwrap_gss_name_t(name2);
  result__c = gss_compare_name(&minor_status__c,name1__c,name2__c,&name_equal__c);
  minor_status = caml_copy_int32(minor_status__c);
  name_equal = Val_bool(name_equal__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = name_equal;
  CAMLreturn(local_0);
}

value net_gss_display_name(value input_name) {
  OM_uint32 minor_status__c;
  gss_name_t input_name__c;
  gss_buffer_t output_name__c;
  gss_OID output_name_type__c;
  OM_uint32 result__c;
  CAMLparam1(input_name);
  CAMLlocal5(minor_status,output_name,output_name_type,result,local_0);
  input_name__c = unwrap_gss_name_t(input_name);
  init_gss_buffer_t(&output_name__c);
  result__c = gss_display_name(&minor_status__c,input_name__c,output_name__c,&output_name_type__c);
  minor_status = caml_copy_int32(minor_status__c);
  output_name = wrap_gss_buffer_t(output_name__c);
  output_name_type = wrap_gss_OID(output_name_type__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(4,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = output_name;
  Field(local_0, 3) = output_name_type;
  CAMLreturn(local_0);
}

value net_gss_import_name(value input_name,value input_name_type) {
  OM_uint32 minor_status__c;
  gss_buffer_t input_name__c;
  gss_OID input_name_type__c;
  gss_name_t output_name__c;
  OM_uint32 result__c;
  CAMLparam2(input_name,input_name_type);
  CAMLlocal4(minor_status,output_name,result,local_0);
  input_name__c = unwrap_gss_buffer_t(input_name);
  input_name_type__c = unwrap_gss_OID(input_name_type);
  result__c = gss_import_name(&minor_status__c,input_name__c,input_name_type__c,&output_name__c);
  minor_status = caml_copy_int32(minor_status__c);
  output_name = wrap_gss_name_t(output_name__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = output_name;
  CAMLreturn(local_0);
}

value net_gss_inquire_cred(value cred) {
  OM_uint32 minor_status__c;
  gss_cred_id_t cred__c;
  gss_name_t name__c;
  OM_uint32 lifetime__c;
  gss_cred_usage_t cred_usage__c;
  gss_OID_set mechanisms__c;
  OM_uint32 result__c;
  CAMLparam1(cred);
  CAMLlocal5(minor_status,name,lifetime,cred_usage,mechanisms);
  CAMLlocal2(result,local_0);
  cred__c = unwrap_gss_cred_id_t(cred);
  result__c = gss_inquire_cred(&minor_status__c,cred__c,&name__c,&lifetime__c,&cred_usage__c,&mechanisms__c);
  minor_status = caml_copy_int32(minor_status__c);
  name = wrap_gss_name_t(name__c);
  lifetime = caml_copy_int32(lifetime__c);
  cred_usage = wrap_gss_cred_usage_t(cred_usage__c);
  mechanisms = wrap_gss_OID_set(mechanisms__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(6,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = name;
  Field(local_0, 3) = lifetime;
  Field(local_0, 4) = cred_usage;
  Field(local_0, 5) = mechanisms;
  CAMLreturn(local_0);
}

value net_gss_inquire_context(value context) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_name_t src_name__c;
  gss_name_t targ_name__c;
  OM_uint32 lifetime_req__c;
  gss_OID mech_type__c;
  flags ctx_flags__c;
  int locally_initiated__c;
  int is_open__c;
  OM_uint32 result__c;
  CAMLparam1(context);
  CAMLlocal5(minor_status,src_name,targ_name,lifetime_req,mech_type);
  CAMLlocal5(ctx_flags,locally_initiated,is_open,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  result__c = gss_inquire_context(&minor_status__c,context__c,&src_name__c,&targ_name__c,&lifetime_req__c,&mech_type__c,&ctx_flags__c,&locally_initiated__c,&is_open__c);
  minor_status = caml_copy_int32(minor_status__c);
  src_name = wrap_gss_name_t(src_name__c);
  targ_name = wrap_gss_name_t(targ_name__c);
  lifetime_req = caml_copy_int32(lifetime_req__c);
  mech_type = wrap_gss_OID(mech_type__c);
  ctx_flags = wrap_flags(ctx_flags__c);
  locally_initiated = Val_bool(locally_initiated__c);
  is_open = Val_bool(is_open__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(9,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = src_name;
  Field(local_0, 3) = targ_name;
  Field(local_0, 4) = lifetime_req;
  Field(local_0, 5) = mech_type;
  Field(local_0, 6) = ctx_flags;
  Field(local_0, 7) = locally_initiated;
  Field(local_0, 8) = is_open;
  CAMLreturn(local_0);
}

value net_gss_wrap_size_limit(value context,value conf_req,value qop_req,value req_output_size) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  int conf_req__c;
  OM_uint32 qop_req__c;
  OM_uint32 req_output_size__c;
  OM_uint32 max_input_size__c;
  OM_uint32 result__c;
  CAMLparam4(context,conf_req,qop_req,req_output_size);
  CAMLlocal4(minor_status,max_input_size,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  conf_req__c = Bool_val(conf_req);
  qop_req__c = Int32_val(qop_req);
  req_output_size__c = Int32_val(req_output_size);
  result__c = gss_wrap_size_limit(&minor_status__c,context__c,conf_req__c,qop_req__c,req_output_size__c,&max_input_size__c);
  minor_status = caml_copy_int32(minor_status__c);
  max_input_size = caml_copy_int32(max_input_size__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = max_input_size;
  CAMLreturn(local_0);
}

value net_gss_add_cred(value input_cred,value desired_name,value desired_mech,value cred_usage,value initiator_time_req,value acceptor_time_req) {
  OM_uint32 minor_status__c;
  gss_cred_id_t input_cred__c;
  gss_name_t desired_name__c;
  gss_OID desired_mech__c;
  gss_cred_usage_t cred_usage__c;
  OM_uint32 initiator_time_req__c;
  OM_uint32 acceptor_time_req__c;
  gss_cred_id_t output_cred__c;
  gss_OID_set actual_mechs__c;
  OM_uint32 initiator_time_rec__c;
  OM_uint32 acceptor_time_rec__c;
  OM_uint32 result__c;
  CAMLparam5(input_cred,desired_name,desired_mech,cred_usage,initiator_time_req);
  CAMLxparam1(acceptor_time_req);
  CAMLlocal5(minor_status,output_cred,actual_mechs,initiator_time_rec,acceptor_time_rec);
  CAMLlocal2(result,local_0);
  input_cred__c = unwrap_gss_cred_id_t(input_cred);
  desired_name__c = unwrap_gss_name_t(desired_name);
  desired_mech__c = unwrap_gss_OID(desired_mech);
  cred_usage__c = unwrap_gss_cred_usage_t(cred_usage);
  initiator_time_req__c = Int32_val(initiator_time_req);
  acceptor_time_req__c = Int32_val(acceptor_time_req);
  caml_enter_blocking_section();
  result__c = gss_add_cred(&minor_status__c,input_cred__c,desired_name__c,desired_mech__c,cred_usage__c,initiator_time_req__c,acceptor_time_req__c,&output_cred__c,&actual_mechs__c,&initiator_time_rec__c,&acceptor_time_rec__c);
  caml_leave_blocking_section();
  minor_status = caml_copy_int32(minor_status__c);
  output_cred = wrap_gss_cred_id_t(output_cred__c);
  actual_mechs = wrap_gss_OID_set(actual_mechs__c);
  initiator_time_rec = caml_copy_int32(initiator_time_rec__c);
  acceptor_time_rec = caml_copy_int32(acceptor_time_rec__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(6,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = output_cred;
  Field(local_0, 3) = actual_mechs;
  Field(local_0, 4) = initiator_time_rec;
  Field(local_0, 5) = acceptor_time_rec;
  CAMLreturn(local_0);
}

value net_gss_add_cred__byte(value * argv, int argn) {
  return net_gss_add_cred(argv[0],argv[1],argv[2],argv[3],argv[4],argv[5]);
}

value net_gss_inquire_cred_by_mech(value cred,value mech_type) {
  OM_uint32 minor_status__c;
  gss_cred_id_t cred__c;
  gss_OID mech_type__c;
  gss_name_t name__c;
  OM_uint32 initiator_lifetime__c;
  OM_uint32 acceptor_lifetime__c;
  gss_cred_usage_t cred_usage__c;
  OM_uint32 result__c;
  CAMLparam2(cred,mech_type);
  CAMLlocal5(minor_status,name,initiator_lifetime,acceptor_lifetime,cred_usage);
  CAMLlocal2(result,local_0);
  cred__c = unwrap_gss_cred_id_t(cred);
  mech_type__c = unwrap_gss_OID(mech_type);
  result__c = gss_inquire_cred_by_mech(&minor_status__c,cred__c,mech_type__c,&name__c,&initiator_lifetime__c,&acceptor_lifetime__c,&cred_usage__c);
  minor_status = caml_copy_int32(minor_status__c);
  name = wrap_gss_name_t(name__c);
  initiator_lifetime = caml_copy_int32(initiator_lifetime__c);
  acceptor_lifetime = caml_copy_int32(acceptor_lifetime__c);
  cred_usage = wrap_gss_cred_usage_t(cred_usage__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(6,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = name;
  Field(local_0, 3) = initiator_lifetime;
  Field(local_0, 4) = acceptor_lifetime;
  Field(local_0, 5) = cred_usage;
  CAMLreturn(local_0);
}

value net_gss_export_sec_context(value context) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_buffer_t interprocess_token__c;
  OM_uint32 result__c;
  CAMLparam1(context);
  CAMLlocal4(minor_status,interprocess_token,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  init_gss_buffer_t(&interprocess_token__c);
  result__c = gss_export_sec_context(&minor_status__c,&context__c,interprocess_token__c);
  if (context__c == GSS_C_NO_CONTEXT) set_gss_ctx_id_t(context, GSS_C_NO_CONTEXT);
  minor_status = caml_copy_int32(minor_status__c);
  interprocess_token = wrap_gss_buffer_t(interprocess_token__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = interprocess_token;
  CAMLreturn(local_0);
}

value net_gss_import_sec_context(value interprocess_token) {
  OM_uint32 minor_status__c;
  gss_buffer_t interprocess_token__c;
  gss_ctx_id_t context__c;
  OM_uint32 result__c;
  CAMLparam1(interprocess_token);
  CAMLlocal4(minor_status,context,result,local_0);
  interprocess_token__c = unwrap_gss_buffer_t(interprocess_token);
  result__c = gss_import_sec_context(&minor_status__c,interprocess_token__c,&context__c);
  minor_status = caml_copy_int32(minor_status__c);
  context = wrap_gss_ctx_id_t(context__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = context;
  CAMLreturn(local_0);
}

value net_gss_inquire_names_for_mech(value mechanism) {
  OM_uint32 minor_status__c;
  gss_OID mechanism__c;
  gss_OID_set name_types__c;
  OM_uint32 result__c;
  CAMLparam1(mechanism);
  CAMLlocal4(minor_status,name_types,result,local_0);
  mechanism__c = unwrap_gss_OID(mechanism);
  result__c = gss_inquire_names_for_mech(&minor_status__c,mechanism__c,&name_types__c);
  minor_status = caml_copy_int32(minor_status__c);
  name_types = wrap_gss_OID_set(name_types__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = name_types;
  CAMLreturn(local_0);
}

value net_gss_inquire_mechs_for_name(value name) {
  OM_uint32 minor_status__c;
  gss_name_t name__c;
  gss_OID_set mech_types__c;
  OM_uint32 result__c;
  CAMLparam1(name);
  CAMLlocal4(minor_status,mech_types,result,local_0);
  name__c = unwrap_gss_name_t(name);
  result__c = gss_inquire_mechs_for_name(&minor_status__c,name__c,&mech_types__c);
  minor_status = caml_copy_int32(minor_status__c);
  mech_types = wrap_gss_OID_set(mech_types__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = mech_types;
  CAMLreturn(local_0);
}

value net_gss_export_name(value name) {
  OM_uint32 minor_status__c;
  gss_name_t name__c;
  gss_buffer_t exported_name__c;
  OM_uint32 result__c;
  CAMLparam1(name);
  CAMLlocal4(minor_status,exported_name,result,local_0);
  name__c = unwrap_gss_name_t(name);
  init_gss_buffer_t(&exported_name__c);
  result__c = gss_export_name(&minor_status__c,name__c,exported_name__c);
  minor_status = caml_copy_int32(minor_status__c);
  exported_name = wrap_gss_buffer_t(exported_name__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = exported_name;
  CAMLreturn(local_0);
}

value net_gss_duplicate_name(value name) {
  OM_uint32 minor_status__c;
  gss_name_t name__c;
  gss_name_t dest_name__c;
  OM_uint32 result__c;
  CAMLparam1(name);
  CAMLlocal4(minor_status,dest_name,result,local_0);
  name__c = unwrap_gss_name_t(name);
  result__c = gss_duplicate_name(&minor_status__c,name__c,&dest_name__c);
  minor_status = caml_copy_int32(minor_status__c);
  dest_name = wrap_gss_name_t(dest_name__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = dest_name;
  CAMLreturn(local_0);
}

value net_gss_canonicalize_name(value input_name,value mech_type) {
  OM_uint32 minor_status__c;
  gss_name_t input_name__c;
  gss_OID mech_type__c;
  gss_name_t output_name__c;
  OM_uint32 result__c;
  CAMLparam2(input_name,mech_type);
  CAMLlocal4(minor_status,output_name,result,local_0);
  input_name__c = unwrap_gss_name_t(input_name);
  mech_type__c = unwrap_gss_OID(mech_type);
  result__c = gss_canonicalize_name(&minor_status__c,input_name__c,mech_type__c,&output_name__c);
  minor_status = caml_copy_int32(minor_status__c);
  output_name = wrap_gss_name_t(output_name__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(3,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  Field(local_0, 2) = output_name;
  CAMLreturn(local_0);
}

value net_gss_delete_sec_context(value context) {
  OM_uint32 minor_status__c;
  gss_ctx_id_t context__c;
  gss_buffer_t output_token__c;
  OM_uint32 result__c;
  CAMLparam1(context);
  CAMLlocal3(minor_status,result,local_0);
  context__c = unwrap_gss_ctx_id_t(context);
  output_token__c = GSS_C_NO_BUFFER;
  result__c = gss_delete_sec_context(&minor_status__c,&context__c,output_token__c);
  if ((result__c & 0xffff0000) == 0) set_gss_ctx_id_t(context, GSS_C_NO_CONTEXT);
  minor_status = caml_copy_int32(minor_status__c);
  result = caml_copy_int32(result__c);
  local_0 = caml_alloc(2,0);
  Field(local_0, 0) = result;
  Field(local_0, 1) = minor_status;
  CAMLreturn(local_0);
}

