#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "caml/mlvalues.h"
#include "caml/alloc.h"
#include "caml/memory.h"
#include "caml/misc.h"
#include "caml/custom.h"
#include "caml/fail.h"
#include "caml/unixsupport.h"
#include "caml/callback.h"
#include "caml/bigarray.h"
#include "caml/threads.h"

static unsigned int uint_val(value v);
static value protected_copy_string(const char *s);

/* This file is included into nettls_nettle_bindings_stubs.c */

#include "./config.h"

#include <nettle/nettle-types.h>
#include <nettle/nettle-meta.h>
#include <nettle/des.h>
#include <nettle/blowfish.h>

#ifdef HAVE_NETTLE_GCM_H
#include <nettle/gcm.h>
#endif

#pragma GCC diagnostic ignored "-Wunused-function"

void nettls_init(void);

typedef struct nettle_cipher *net_nettle_cipher_t;
typedef void *net_nettle_cipher_ctx_t;
typedef void *net_nettle_gcm_aes_ctx_t;

typedef struct nettle_hash *net_nettle_hash_t;
typedef void *net_nettle_hash_ctx_t;

#define raise_null_pointer net_nettle_null_pointer

static void net_nettle_null_pointer(void) {
    caml_raise_constant(*caml_named_value
                          ("Nettls_nettle_bindings.Null_pointer"));
}

#ifndef HAVE_TY_nettle_cipher_func
#define nettle_cipher_func nettle_crypt_func
#endif


/* Generic API */

static net_nettle_cipher_ctx_t
         net_nettle_create_cipher_ctx(net_nettle_cipher_t cipher) {
    void *p;
    p = stat_alloc(cipher->context_size);
    return p;
}

static void net_nettle_free(void *ctx) {
    stat_free(ctx);
}

static void net_nettle_destroy_cipher(net_nettle_cipher_t cipher) {
    /* do nothing, ciphers are non-destructable */
}

static void net_nettle_set_encrypt_key(net_nettle_cipher_t cipher,
                                       net_nettle_cipher_ctx_t ctx,
                                       unsigned int length,
                                       const uint8_t *key) {
#ifdef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
    if (length != cipher->key_size)
        failwith("net_nettl_set_encrypt_key: key has wrong size");
    cipher->set_encrypt_key(ctx, key);
#else
    cipher->set_encrypt_key(ctx, length, key);
#endif
}


static void net_nettle_set_decrypt_key(net_nettle_cipher_t cipher,
                                       net_nettle_cipher_ctx_t ctx,
                                       unsigned int length,
                                       const uint8_t *key) {
#ifdef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
    if (length != cipher->key_size)
        failwith("net_nettl_set_decrypt_key: key has wrong size");
    cipher->set_decrypt_key(ctx, key);
#else
    cipher->set_decrypt_key(ctx, length, key);
#endif
}

static void net_nettle_encrypt(net_nettle_cipher_t cipher,
                               net_nettle_cipher_ctx_t ctx,
                               unsigned int length,
                               uint8_t *dst,
                               const uint8_t *src) {
    cipher->encrypt(ctx, length, dst, src);
}

static void net_nettle_decrypt(net_nettle_cipher_t cipher,
                               net_nettle_cipher_ctx_t ctx,
                               unsigned int length,
                               uint8_t *dst,
                               const uint8_t *src) {
    cipher->decrypt(ctx, length, dst, src);
}

static const char *net_nettle_cipher_name(net_nettle_cipher_t cipher) {
    return cipher->name;
}

#ifndef HAVE_FUN_nettle_get_ciphers
#ifndef HAVE_FUN_nettle_ciphers
const struct nettle_cipher * const nettle_ciphers[] = {
  &nettle_aes128,
  &nettle_aes192,
  &nettle_aes256,
#ifdef HAVE_NETTLE_CAMELLIA_H
  &nettle_camellia128,
  &nettle_camellia192,
  &nettle_camellia256,
#endif
  &nettle_cast128,
  &nettle_serpent128,
  &nettle_serpent192,
  &nettle_serpent256,
  &nettle_twofish128,
  &nettle_twofish192,
  &nettle_twofish256,
  &nettle_arctwo40,
  &nettle_arctwo64,
  &nettle_arctwo128,
  &nettle_arctwo_gutmann128,
  NULL
};
#endif
#endif

static void net_nettle_ciphers(net_nettle_cipher_t **ciphers,
                               size_t *n) {
    size_t k;
    const struct nettle_cipher * const *nciphers;
#ifdef HAVE_FUN_nettle_get_ciphers
    nciphers = nettle_get_ciphers();
#else
    nciphers = nettle_ciphers;
#endif
    k = 0;
    while (nciphers[k] != NULL) k++;
    *ciphers = (net_nettle_cipher_t *) nettle_ciphers;
    *n = k;
}


/* Extensions to the generic API */

#ifdef  HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
#define MAYBE_LENGTH
#else
#define MAYBE_LENGTH unsigned int length,
#endif


static void net_des_set_key(void *ctx, MAYBE_LENGTH
                            const uint8_t *key) {
    struct des_ctx *dctx;
    dctx = (struct des_ctx *) ctx;
    des_set_key(dctx, key);
}

static void net_des_encrypt(void *ctx, unsigned int length,
                            uint8_t *dst, const uint8_t *src) {
    des_encrypt((struct des_ctx *) ctx, length, dst, src);
}

static void net_des_decrypt(void *ctx, unsigned int length,
                            uint8_t *dst, const uint8_t *src) {
    des_decrypt((struct des_ctx *) ctx, length, dst, src);
}

static const struct nettle_cipher net_nettle_des =
  { .name = "des",
    .context_size = sizeof(struct des_ctx),
    .block_size = 64,
    .key_size = 56,
    .set_encrypt_key = net_des_set_key,
    .set_decrypt_key = net_des_set_key,
    .encrypt = (nettle_cipher_func *) net_des_encrypt,
    .decrypt = (nettle_cipher_func *) net_des_decrypt
  };

static void net_des3_set_key(void *ctx, MAYBE_LENGTH
                             const uint8_t *key) {
    struct des3_ctx *dctx;
    dctx = (struct des3_ctx *) ctx;
    des3_set_key(dctx, key);
}

static void net_des3_encrypt(void *ctx, unsigned int length,
                            uint8_t *dst, const uint8_t *src) {
    des3_encrypt((struct des3_ctx *) ctx, length, dst, src);
}

static void net_des3_decrypt(void *ctx, unsigned int length,
                            uint8_t *dst, const uint8_t *src) {
    des3_decrypt((struct des3_ctx *) ctx, length, dst, src);
}

static const struct nettle_cipher net_nettle_des3 =
  { .name = "des3",
    .context_size = sizeof(struct des3_ctx),
    .block_size = 64,
    .key_size = 112,
    .set_encrypt_key = net_des3_set_key,
    .set_decrypt_key = net_des3_set_key,
    .encrypt = (nettle_cipher_func *) net_des3_encrypt,
    .decrypt = (nettle_cipher_func *) net_des3_decrypt
  };


/* Blowfish: this cipher has a variable length. Older versions of Nettle
   support this well, but newer nettle_cipher types do not, so we have to
   restrict to blowfish128
*/


#ifndef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
static void net_blowfish_set_key(void *ctx, unsigned int length,
                                 const uint8_t *key) {
    struct blowfish_ctx *dctx;
    dctx = (struct blowfish_ctx *) ctx;
    blowfish_set_key(dctx, length, key);
}
#endif

#ifdef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
static void net_blowfish128_set_key(void *ctx, const uint8_t *key) {
    struct blowfish_ctx *dctx;
    dctx = (struct blowfish_ctx *) ctx;
    blowfish128_set_key(dctx, key);
}
#endif

static void net_blowfish_encrypt(void *ctx, unsigned int length,
                                 uint8_t *dst, const uint8_t *src) {
    blowfish_encrypt((struct blowfish_ctx *) ctx, length, dst, src);
}

static void net_blowfish_decrypt(void *ctx, unsigned int length,
                                uint8_t *dst, const uint8_t *src) {
    blowfish_decrypt((struct blowfish_ctx *) ctx, length, dst, src);
}

#ifndef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
static const struct nettle_cipher net_nettle_blowfish =
  { .name = "blowfish",
    .context_size = sizeof(struct blowfish_ctx),
    .block_size = 64,
    .key_size = 16,  /* variable */
    .set_encrypt_key = net_blowfish_set_key,
    .set_decrypt_key = net_blowfish_set_key,
    .encrypt = (nettle_cipher_func *) net_blowfish_encrypt,
    .decrypt = (nettle_cipher_func *) net_blowfish_decrypt
  };
#endif


#ifdef HAVE_NETTLE_SET_KEY_WITH_TWO_ARGS
static const struct nettle_cipher net_nettle_blowfish =
  { .name = "blowfish128",
    .context_size = sizeof(struct blowfish_ctx),
    .block_size = 64,
    .key_size = 16,  /* variable */
    .set_encrypt_key = net_blowfish128_set_key,
    .set_decrypt_key = net_blowfish128_set_key,
    .encrypt = (nettle_cipher_func *) net_blowfish_encrypt,
    .decrypt = (nettle_cipher_func *) net_blowfish_decrypt
  };
#endif

static const struct nettle_cipher * const ext_ciphers[] = {
    &net_nettle_des,
    &net_nettle_des3,
    &net_nettle_blowfish,
    NULL
};

static void net_ext_ciphers(net_nettle_cipher_t **ciphers,
                            size_t *n) {
    size_t k;
    k = 0;
    while (ext_ciphers[k] != NULL) k++;
    *ciphers = (net_nettle_cipher_t *) ext_ciphers;
    *n = k;
}


/* GCM */

/* TODO: newer versions of Nettle have a nettle_aead abstraction in
   nettle-meta.h
*/

static net_nettle_gcm_aes_ctx_t net_nettle_gcm_aes_init(void) {
#ifdef HAVE_NETTLE_GCM_H
    return stat_alloc(sizeof(struct gcm_aes_ctx));
#else
    return NULL;
#endif
}


static int net_have_gcm_aes(void) {
#ifdef HAVE_NETTLE_GCM_H
    return 1;
#else
    return 0;
#endif
}


/* Hashes */

static void net_nettle_destroy_hash(net_nettle_hash_t hash) {
    /* do nothing, hashes are non-destructable */
}

static const char *net_nettle_hash_name(net_nettle_hash_t hash) {
    return hash->name;
}

static net_nettle_hash_ctx_t
         net_nettle_create_hash_ctx(net_nettle_hash_t hash) {
    void *p;
    p = stat_alloc(hash->context_size);
    return p;
}

static void net_nettle_hash_init(net_nettle_hash_t hash,
                                 net_nettle_hash_ctx_t ctx) {
    hash->init(ctx);
}

static void net_nettle_hash_update(net_nettle_hash_t hash,
                                   net_nettle_hash_ctx_t ctx,
                                   unsigned int length,
                                   const uint8_t *src) {
    hash->update(ctx, length, src);
}

static void net_nettle_hash_digest(net_nettle_hash_t hash,
                                   net_nettle_hash_ctx_t ctx,
                                   unsigned int length,
                                   uint8_t *dst) {
    hash->digest(ctx, length, dst);
}

#ifndef HAVE_FUN_nettle_get_hashes
#ifndef HAVE_FUN_nettle_hashes
const struct nettle_hash * const nettle_hashes[] = {
    &nettle_md2,
    &nettle_md4,
    &nettle_md5,
    &nettle_sha1,
    &nettle_sha256,
    NULL
};
#endif
#endif

static void net_nettle_hashes(net_nettle_hash_t **hashes,
                              size_t *n) {
    size_t k;
    const struct nettle_hash * const *nhashes;
#ifdef HAVE_FUN_nettle_get_hashes
    nhashes = nettle_get_hashes();
#else
    nhashes = nettle_hashes;
#endif
    k = 0;
    while (nettle_hashes[k] != NULL) k++;
    *hashes = (net_nettle_hash_t *) nettle_hashes;
    *n = k;
}
static unsigned int uint_val(value v) {
    if (Int_val(v) < 0) invalid_argument("negative integer");
    return (unsigned int) Int_val(v);
}

static value protected_copy_string(const char *s) {
    if (s==NULL) raise_null_pointer();
    return caml_copy_string(s);
}

/************** net_nettle_cipher_t *************/

struct absstruct_net_nettle_cipher_t { net_nettle_cipher_t value; long tag; long oid; };

#define absstructptr_net_nettle_cipher_t_val(v) ((struct absstruct_net_nettle_cipher_t *) (Data_custom_val(v)))
#define abs_net_nettle_cipher_t_unwrap(v) (absstructptr_net_nettle_cipher_t_val(v)->value)
#define abs_net_nettle_cipher_t_tag(v) (absstructptr_net_nettle_cipher_t_val(v)->tag)
long abs_net_nettle_cipher_t_oid = 0;

static int abs_net_nettle_cipher_t_compare(value v1, value v2) {
  struct absstruct_net_nettle_cipher_t *p1;
  struct absstruct_net_nettle_cipher_t *p2;
  p1 = absstructptr_net_nettle_cipher_t_val(v1);
  p2 = absstructptr_net_nettle_cipher_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_net_nettle_cipher_t_finalize(value v1) {
  struct absstruct_net_nettle_cipher_t *p1;
  p1 = absstructptr_net_nettle_cipher_t_val(v1);
  net_nettle_destroy_cipher(p1->value);
}

static struct custom_operations abs_net_nettle_cipher_t_ops = {
  "",
  abs_net_nettle_cipher_t_finalize,
  abs_net_nettle_cipher_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static net_nettle_cipher_t unwrap_net_nettle_cipher_t(value v) {
  net_nettle_cipher_t r;
  r = abs_net_nettle_cipher_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_net_nettle_cipher_t(value v) {
  return abs_net_nettle_cipher_t_tag(Field(v,0));
}

static value twrap_net_nettle_cipher_t(long tag, net_nettle_cipher_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_net_nettle_cipher_t: NULL pointer");
  v = caml_alloc_custom(&abs_net_nettle_cipher_t_ops, sizeof(struct absstruct_net_nettle_cipher_t), 0, 1);
  absstructptr_net_nettle_cipher_t_val(v)->tag = tag;
  absstructptr_net_nettle_cipher_t_val(v)->value = x;
  absstructptr_net_nettle_cipher_t_val(v)->oid = abs_net_nettle_cipher_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_net_nettle_cipher_t(net_nettle_cipher_t x) {
  return twrap_net_nettle_cipher_t(0, x);
}

static void attach_net_nettle_cipher_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** net_nettle_cipher_ctx_t *************/

struct absstruct_net_nettle_cipher_ctx_t { net_nettle_cipher_ctx_t value; long tag; long oid; };

#define absstructptr_net_nettle_cipher_ctx_t_val(v) ((struct absstruct_net_nettle_cipher_ctx_t *) (Data_custom_val(v)))
#define abs_net_nettle_cipher_ctx_t_unwrap(v) (absstructptr_net_nettle_cipher_ctx_t_val(v)->value)
#define abs_net_nettle_cipher_ctx_t_tag(v) (absstructptr_net_nettle_cipher_ctx_t_val(v)->tag)
long abs_net_nettle_cipher_ctx_t_oid = 0;

static int abs_net_nettle_cipher_ctx_t_compare(value v1, value v2) {
  struct absstruct_net_nettle_cipher_ctx_t *p1;
  struct absstruct_net_nettle_cipher_ctx_t *p2;
  p1 = absstructptr_net_nettle_cipher_ctx_t_val(v1);
  p2 = absstructptr_net_nettle_cipher_ctx_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_net_nettle_cipher_ctx_t_finalize(value v1) {
  struct absstruct_net_nettle_cipher_ctx_t *p1;
  p1 = absstructptr_net_nettle_cipher_ctx_t_val(v1);
  net_nettle_free(p1->value);
}

static struct custom_operations abs_net_nettle_cipher_ctx_t_ops = {
  "",
  abs_net_nettle_cipher_ctx_t_finalize,
  abs_net_nettle_cipher_ctx_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static net_nettle_cipher_ctx_t unwrap_net_nettle_cipher_ctx_t(value v) {
  net_nettle_cipher_ctx_t r;
  r = abs_net_nettle_cipher_ctx_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_net_nettle_cipher_ctx_t(value v) {
  return abs_net_nettle_cipher_ctx_t_tag(Field(v,0));
}

static value twrap_net_nettle_cipher_ctx_t(long tag, net_nettle_cipher_ctx_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_net_nettle_cipher_ctx_t: NULL pointer");
  v = caml_alloc_custom(&abs_net_nettle_cipher_ctx_t_ops, sizeof(struct absstruct_net_nettle_cipher_ctx_t), 0, 1);
  absstructptr_net_nettle_cipher_ctx_t_val(v)->tag = tag;
  absstructptr_net_nettle_cipher_ctx_t_val(v)->value = x;
  absstructptr_net_nettle_cipher_ctx_t_val(v)->oid = abs_net_nettle_cipher_ctx_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_net_nettle_cipher_ctx_t(net_nettle_cipher_ctx_t x) {
  return twrap_net_nettle_cipher_ctx_t(0, x);
}

static void attach_net_nettle_cipher_ctx_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** net_nettle_gcm_aes_ctx_t *************/

struct absstruct_net_nettle_gcm_aes_ctx_t { net_nettle_gcm_aes_ctx_t value; long tag; long oid; };

#define absstructptr_net_nettle_gcm_aes_ctx_t_val(v) ((struct absstruct_net_nettle_gcm_aes_ctx_t *) (Data_custom_val(v)))
#define abs_net_nettle_gcm_aes_ctx_t_unwrap(v) (absstructptr_net_nettle_gcm_aes_ctx_t_val(v)->value)
#define abs_net_nettle_gcm_aes_ctx_t_tag(v) (absstructptr_net_nettle_gcm_aes_ctx_t_val(v)->tag)
long abs_net_nettle_gcm_aes_ctx_t_oid = 0;

static int abs_net_nettle_gcm_aes_ctx_t_compare(value v1, value v2) {
  struct absstruct_net_nettle_gcm_aes_ctx_t *p1;
  struct absstruct_net_nettle_gcm_aes_ctx_t *p2;
  p1 = absstructptr_net_nettle_gcm_aes_ctx_t_val(v1);
  p2 = absstructptr_net_nettle_gcm_aes_ctx_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_net_nettle_gcm_aes_ctx_t_finalize(value v1) {
  struct absstruct_net_nettle_gcm_aes_ctx_t *p1;
  p1 = absstructptr_net_nettle_gcm_aes_ctx_t_val(v1);
  net_nettle_free(p1->value);
}

static struct custom_operations abs_net_nettle_gcm_aes_ctx_t_ops = {
  "",
  abs_net_nettle_gcm_aes_ctx_t_finalize,
  abs_net_nettle_gcm_aes_ctx_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static net_nettle_gcm_aes_ctx_t unwrap_net_nettle_gcm_aes_ctx_t(value v) {
  net_nettle_gcm_aes_ctx_t r;
  r = abs_net_nettle_gcm_aes_ctx_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_net_nettle_gcm_aes_ctx_t(value v) {
  return abs_net_nettle_gcm_aes_ctx_t_tag(Field(v,0));
}

static value twrap_net_nettle_gcm_aes_ctx_t(long tag, net_nettle_gcm_aes_ctx_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_net_nettle_gcm_aes_ctx_t: NULL pointer");
  v = caml_alloc_custom(&abs_net_nettle_gcm_aes_ctx_t_ops, sizeof(struct absstruct_net_nettle_gcm_aes_ctx_t), 0, 1);
  absstructptr_net_nettle_gcm_aes_ctx_t_val(v)->tag = tag;
  absstructptr_net_nettle_gcm_aes_ctx_t_val(v)->value = x;
  absstructptr_net_nettle_gcm_aes_ctx_t_val(v)->oid = abs_net_nettle_gcm_aes_ctx_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_net_nettle_gcm_aes_ctx_t(net_nettle_gcm_aes_ctx_t x) {
  return twrap_net_nettle_gcm_aes_ctx_t(0, x);
}

static void attach_net_nettle_gcm_aes_ctx_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** net_nettle_hash_t *************/

struct absstruct_net_nettle_hash_t { net_nettle_hash_t value; long tag; long oid; };

#define absstructptr_net_nettle_hash_t_val(v) ((struct absstruct_net_nettle_hash_t *) (Data_custom_val(v)))
#define abs_net_nettle_hash_t_unwrap(v) (absstructptr_net_nettle_hash_t_val(v)->value)
#define abs_net_nettle_hash_t_tag(v) (absstructptr_net_nettle_hash_t_val(v)->tag)
long abs_net_nettle_hash_t_oid = 0;

static int abs_net_nettle_hash_t_compare(value v1, value v2) {
  struct absstruct_net_nettle_hash_t *p1;
  struct absstruct_net_nettle_hash_t *p2;
  p1 = absstructptr_net_nettle_hash_t_val(v1);
  p2 = absstructptr_net_nettle_hash_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_net_nettle_hash_t_finalize(value v1) {
  struct absstruct_net_nettle_hash_t *p1;
  p1 = absstructptr_net_nettle_hash_t_val(v1);
  net_nettle_destroy_hash(p1->value);
}

static struct custom_operations abs_net_nettle_hash_t_ops = {
  "",
  abs_net_nettle_hash_t_finalize,
  abs_net_nettle_hash_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static net_nettle_hash_t unwrap_net_nettle_hash_t(value v) {
  net_nettle_hash_t r;
  r = abs_net_nettle_hash_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_net_nettle_hash_t(value v) {
  return abs_net_nettle_hash_t_tag(Field(v,0));
}

static value twrap_net_nettle_hash_t(long tag, net_nettle_hash_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_net_nettle_hash_t: NULL pointer");
  v = caml_alloc_custom(&abs_net_nettle_hash_t_ops, sizeof(struct absstruct_net_nettle_hash_t), 0, 1);
  absstructptr_net_nettle_hash_t_val(v)->tag = tag;
  absstructptr_net_nettle_hash_t_val(v)->value = x;
  absstructptr_net_nettle_hash_t_val(v)->oid = abs_net_nettle_hash_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_net_nettle_hash_t(net_nettle_hash_t x) {
  return twrap_net_nettle_hash_t(0, x);
}

static void attach_net_nettle_hash_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

/************** net_nettle_hash_ctx_t *************/

struct absstruct_net_nettle_hash_ctx_t { net_nettle_hash_ctx_t value; long tag; long oid; };

#define absstructptr_net_nettle_hash_ctx_t_val(v) ((struct absstruct_net_nettle_hash_ctx_t *) (Data_custom_val(v)))
#define abs_net_nettle_hash_ctx_t_unwrap(v) (absstructptr_net_nettle_hash_ctx_t_val(v)->value)
#define abs_net_nettle_hash_ctx_t_tag(v) (absstructptr_net_nettle_hash_ctx_t_val(v)->tag)
long abs_net_nettle_hash_ctx_t_oid = 0;

static int abs_net_nettle_hash_ctx_t_compare(value v1, value v2) {
  struct absstruct_net_nettle_hash_ctx_t *p1;
  struct absstruct_net_nettle_hash_ctx_t *p2;
  p1 = absstructptr_net_nettle_hash_ctx_t_val(v1);
  p2 = absstructptr_net_nettle_hash_ctx_t_val(v2);
  return p1->oid - p2->oid;
}

static void abs_net_nettle_hash_ctx_t_finalize(value v1) {
  struct absstruct_net_nettle_hash_ctx_t *p1;
  p1 = absstructptr_net_nettle_hash_ctx_t_val(v1);
  net_nettle_free(p1->value);
}

static struct custom_operations abs_net_nettle_hash_ctx_t_ops = {
  "",
  abs_net_nettle_hash_ctx_t_finalize,
  abs_net_nettle_hash_ctx_t_compare,
  custom_hash_default,
  custom_serialize_default,
  custom_deserialize_default
};

static net_nettle_hash_ctx_t unwrap_net_nettle_hash_ctx_t(value v) {
  net_nettle_hash_ctx_t r;
  r = abs_net_nettle_hash_ctx_t_unwrap(Field(v,0));
  if (r == NULL) raise_null_pointer();
  return r;
}

static long tag_net_nettle_hash_ctx_t(value v) {
  return abs_net_nettle_hash_ctx_t_tag(Field(v,0));
}

static value twrap_net_nettle_hash_ctx_t(long tag, net_nettle_hash_ctx_t x) {
  CAMLparam0();
  CAMLlocal2(v,r);
  if (x == NULL) failwith("wrap_net_nettle_hash_ctx_t: NULL pointer");
  v = caml_alloc_custom(&abs_net_nettle_hash_ctx_t_ops, sizeof(struct absstruct_net_nettle_hash_ctx_t), 0, 1);
  absstructptr_net_nettle_hash_ctx_t_val(v)->tag = tag;
  absstructptr_net_nettle_hash_ctx_t_val(v)->value = x;
  absstructptr_net_nettle_hash_ctx_t_val(v)->oid = abs_net_nettle_hash_ctx_t_oid++;
  r = caml_alloc(2,0);
  Field(r,0) = v;
  Field(r,1) = Val_int(0);
  CAMLreturn(r);
}

static value wrap_net_nettle_hash_ctx_t(net_nettle_hash_ctx_t x) {
  return twrap_net_nettle_hash_ctx_t(0, x);
}

static void attach_net_nettle_hash_ctx_t(value v, value aux) {
  CAMLparam2(v,aux);
  CAMLlocal1(h);
  h = caml_alloc(2,0);
  Field(h,0) = aux;
  Field(h,1) = Field(v,1);
  Store_field(v,1,h);
  CAMLreturn0;
}

value net_net_nettle_ciphers(value dummy) {
  net_nettle_cipher_t * cipher_list__c;
  long local_0;
  size_t cipher_list_size__c;
  CAMLparam1(dummy);
  CAMLlocal1(cipher_list);
  nettls_init();
  net_nettle_ciphers(&cipher_list__c,&cipher_list_size__c);
  if (cipher_list__c == NULL)
    cipher_list = caml_alloc(0,0);
  else {
    cipher_list = caml_alloc(cipher_list_size__c,0);
    for (local_0 = 0; local_0 < cipher_list_size__c; local_0++) {
      Store_field(cipher_list, local_0, wrap_net_nettle_cipher_t(cipher_list__c[local_0]));
    };
  };
  CAMLreturn(cipher_list);
}

value net_net_ext_ciphers(value dummy) {
  net_nettle_cipher_t * cipher_list__c;
  long local_0;
  size_t cipher_list_size__c;
  CAMLparam1(dummy);
  CAMLlocal1(cipher_list);
  nettls_init();
  net_ext_ciphers(&cipher_list__c,&cipher_list_size__c);
  if (cipher_list__c == NULL)
    cipher_list = caml_alloc(0,0);
  else {
    cipher_list = caml_alloc(cipher_list_size__c,0);
    for (local_0 = 0; local_0 < cipher_list_size__c; local_0++) {
      Store_field(cipher_list, local_0, wrap_net_nettle_cipher_t(cipher_list__c[local_0]));
    };
  };
  CAMLreturn(cipher_list);
}

value net_net_nettle_cipher_name(value cipher) {
  net_nettle_cipher_t cipher__c;
  const char * result__c;
  CAMLparam1(cipher);
  CAMLlocal1(result);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  nettls_init();
  result__c = net_nettle_cipher_name(cipher__c);
  result = protected_copy_string(result__c);
  CAMLreturn(result);
}

value net_net_nettle_create_cipher_ctx(value cipher) {
  net_nettle_cipher_t cipher__c;
  net_nettle_cipher_ctx_t result__c;
  CAMLparam1(cipher);
  CAMLlocal1(result);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  nettls_init();
  result__c = net_nettle_create_cipher_ctx(cipher__c);
  result = wrap_net_nettle_cipher_ctx_t(result__c);
  CAMLreturn(result);
}

value net_net_nettle_set_encrypt_key(value cipher,value ctx,value key) {
  net_nettle_cipher_t cipher__c;
  net_nettle_cipher_ctx_t ctx__c;
  size_t key_length__c;
  void * key__c;
  CAMLparam3(cipher,ctx,key);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  ctx__c = unwrap_net_nettle_cipher_ctx_t(ctx);
  key_length__c = caml_string_length(key);
  key__c = String_val(key);
  nettls_init();
  net_nettle_set_encrypt_key(cipher__c,ctx__c,key_length__c,key__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_set_decrypt_key(value cipher,value ctx,value key) {
  net_nettle_cipher_t cipher__c;
  net_nettle_cipher_ctx_t ctx__c;
  size_t key_length__c;
  void * key__c;
  CAMLparam3(cipher,ctx,key);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  ctx__c = unwrap_net_nettle_cipher_ctx_t(ctx);
  key_length__c = caml_string_length(key);
  key__c = String_val(key);
  nettls_init();
  net_nettle_set_decrypt_key(cipher__c,ctx__c,key_length__c,key__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_encrypt(value cipher,value ctx,value length,value dst,value src) {
  net_nettle_cipher_t cipher__c;
  net_nettle_cipher_ctx_t ctx__c;
  unsigned int length__c;
  void * dst__c;
  void * src__c;
  CAMLparam5(cipher,ctx,length,dst,src);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  ctx__c = unwrap_net_nettle_cipher_ctx_t(ctx);
  length__c = uint_val(length);
  dst__c = Caml_ba_data_val(dst);
  src__c = Caml_ba_data_val(src);
  nettls_init();
  net_nettle_encrypt(cipher__c,ctx__c,length__c,dst__c,src__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_decrypt(value cipher,value ctx,value length,value dst,value src) {
  net_nettle_cipher_t cipher__c;
  net_nettle_cipher_ctx_t ctx__c;
  unsigned int length__c;
  void * dst__c;
  void * src__c;
  CAMLparam5(cipher,ctx,length,dst,src);
  cipher__c = unwrap_net_nettle_cipher_t(cipher);
  ctx__c = unwrap_net_nettle_cipher_ctx_t(ctx);
  length__c = uint_val(length);
  dst__c = Caml_ba_data_val(dst);
  src__c = Caml_ba_data_val(src);
  nettls_init();
  net_nettle_decrypt(cipher__c,ctx__c,length__c,dst__c,src__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_gcm_aes_init(value dummy) {
  net_nettle_gcm_aes_ctx_t result__c;
  CAMLparam1(dummy);
  CAMLlocal1(result);
  nettls_init();
  result__c = net_nettle_gcm_aes_init();
  result = wrap_net_nettle_gcm_aes_ctx_t(result__c);
  CAMLreturn(result);
}

value net_nettle_gcm_aes_set_key(value ctx,value key) {
#ifdef HAVE_FUN_nettle_gcm_aes_set_key
  net_nettle_gcm_aes_ctx_t ctx__c;
  size_t key_length__c;
  void * key__c;
  CAMLparam2(ctx,key);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  key_length__c = caml_string_length(key);
  key__c = String_val(key);
  nettls_init();
  nettle_gcm_aes_set_key(ctx__c,key_length__c,key__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_set_key");
#endif
}

value net_nettle_gcm_aes_set_iv(value ctx,value iv) {
#ifdef HAVE_FUN_nettle_gcm_aes_set_iv
  net_nettle_gcm_aes_ctx_t ctx__c;
  size_t iv_length__c;
  void * iv__c;
  CAMLparam2(ctx,iv);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  iv_length__c = caml_string_length(iv);
  iv__c = String_val(iv);
  nettls_init();
  nettle_gcm_aes_set_iv(ctx__c,iv_length__c,iv__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_set_iv");
#endif
}

value net_nettle_gcm_aes_update(value ctx,value data) {
#ifdef HAVE_FUN_nettle_gcm_aes_update
  net_nettle_gcm_aes_ctx_t ctx__c;
  size_t data_length__c;
  void * data__c;
  CAMLparam2(ctx,data);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  data_length__c = caml_string_length(data);
  data__c = String_val(data);
  nettls_init();
  nettle_gcm_aes_update(ctx__c,data_length__c,data__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_update");
#endif
}

value net_nettle_gcm_aes_encrypt(value ctx,value length,value dst,value src) {
#ifdef HAVE_FUN_nettle_gcm_aes_encrypt
  net_nettle_gcm_aes_ctx_t ctx__c;
  unsigned int length__c;
  void * dst__c;
  void * src__c;
  CAMLparam4(ctx,length,dst,src);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  length__c = uint_val(length);
  dst__c = Caml_ba_data_val(dst);
  src__c = Caml_ba_data_val(src);
  nettls_init();
  nettle_gcm_aes_encrypt(ctx__c,length__c,dst__c,src__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_encrypt");
#endif
}

value net_nettle_gcm_aes_decrypt(value ctx,value length,value dst,value src) {
#ifdef HAVE_FUN_nettle_gcm_aes_decrypt
  net_nettle_gcm_aes_ctx_t ctx__c;
  unsigned int length__c;
  void * dst__c;
  void * src__c;
  CAMLparam4(ctx,length,dst,src);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  length__c = uint_val(length);
  dst__c = Caml_ba_data_val(dst);
  src__c = Caml_ba_data_val(src);
  nettls_init();
  nettle_gcm_aes_decrypt(ctx__c,length__c,dst__c,src__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_decrypt");
#endif
}

value net_nettle_gcm_aes_digest(value ctx,value mac) {
#ifdef HAVE_FUN_nettle_gcm_aes_digest
  net_nettle_gcm_aes_ctx_t ctx__c;
  size_t length__c;
  void * mac__c;
  CAMLparam2(ctx,mac);
  ctx__c = unwrap_net_nettle_gcm_aes_ctx_t(ctx);
  length__c = caml_string_length(mac);
  mac__c = String_val(mac);
  nettls_init();
  nettle_gcm_aes_digest(ctx__c,length__c,mac__c);
  CAMLreturn(Val_unit);
#else
  invalid_argument("nettle_gcm_aes_digest");
#endif
}

value net_net_have_gcm_aes(value dummy) {
  int result__c;
  CAMLparam1(dummy);
  CAMLlocal1(result);
  nettls_init();
  result__c = net_have_gcm_aes();
  result = Val_bool(result__c);
  CAMLreturn(result);
}

value net_net_nettle_hashes(value dummy) {
  net_nettle_hash_t * hash_list__c;
  long local_0;
  size_t hash_list_size__c;
  CAMLparam1(dummy);
  CAMLlocal1(hash_list);
  nettls_init();
  net_nettle_hashes(&hash_list__c,&hash_list_size__c);
  if (hash_list__c == NULL)
    hash_list = caml_alloc(0,0);
  else {
    hash_list = caml_alloc(hash_list_size__c,0);
    for (local_0 = 0; local_0 < hash_list_size__c; local_0++) {
      Store_field(hash_list, local_0, wrap_net_nettle_hash_t(hash_list__c[local_0]));
    };
  };
  CAMLreturn(hash_list);
}

value net_net_nettle_hash_name(value hash) {
  net_nettle_hash_t hash__c;
  const char * result__c;
  CAMLparam1(hash);
  CAMLlocal1(result);
  hash__c = unwrap_net_nettle_hash_t(hash);
  nettls_init();
  result__c = net_nettle_hash_name(hash__c);
  result = protected_copy_string(result__c);
  CAMLreturn(result);
}

value net_net_nettle_create_hash_ctx(value hash) {
  net_nettle_hash_t hash__c;
  net_nettle_hash_ctx_t result__c;
  CAMLparam1(hash);
  CAMLlocal1(result);
  hash__c = unwrap_net_nettle_hash_t(hash);
  nettls_init();
  result__c = net_nettle_create_hash_ctx(hash__c);
  result = wrap_net_nettle_hash_ctx_t(result__c);
  CAMLreturn(result);
}

value net_net_nettle_hash_init(value hash,value ctx) {
  net_nettle_hash_t hash__c;
  net_nettle_hash_ctx_t ctx__c;
  CAMLparam2(hash,ctx);
  hash__c = unwrap_net_nettle_hash_t(hash);
  ctx__c = unwrap_net_nettle_hash_ctx_t(ctx);
  nettls_init();
  net_nettle_hash_init(hash__c,ctx__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_hash_update(value hash,value ctx,value src) {
  net_nettle_hash_t hash__c;
  net_nettle_hash_ctx_t ctx__c;
  size_t length__c;
  void * src__c;
  CAMLparam3(hash,ctx,src);
  hash__c = unwrap_net_nettle_hash_t(hash);
  ctx__c = unwrap_net_nettle_hash_ctx_t(ctx);
  length__c = caml_ba_byte_size(Caml_ba_array_val(src));
  src__c = Caml_ba_data_val(src);
  nettls_init();
  net_nettle_hash_update(hash__c,ctx__c,length__c,src__c);
  CAMLreturn(Val_unit);
}

value net_net_nettle_hash_digest(value hash,value ctx,value src) {
  net_nettle_hash_t hash__c;
  net_nettle_hash_ctx_t ctx__c;
  size_t length__c;
  void * src__c;
  CAMLparam3(hash,ctx,src);
  hash__c = unwrap_net_nettle_hash_t(hash);
  ctx__c = unwrap_net_nettle_hash_ctx_t(ctx);
  length__c = caml_string_length(src);
  src__c = String_val(src);
  nettls_init();
  net_nettle_hash_digest(hash__c,ctx__c,length__c,src__c);
  CAMLreturn(Val_unit);
}

